import os
from pathlib import Path
import shutil
import click
import pandas as pd
import scales_nlp
from scales_nlp import config


@click.command()
@click.option('--push/--no-push', default=False, help='Push project to PyPI')
def build(push):
    """
    Build and publish the project to PyPI.
    """
    repo_dir = Path(scales_nlp.__file__).parents[2]
    os.system(f'python {repo_dir}/setup.py bdist_wheel --universal')
    if push:
        os.system('twine upload dist/*')


@click.command()
@click.option('--build/--no-build', default=False, help='Push project to PyPI')
def docs(build):
    """
    Start a local server and optionally build the docs with `--build`.
    """
    docs_dir = Path(scales_nlp.__file__).parents[2] / 'docs'
    build_dir = docs_dir / 'build' / 'html'
    if build:
        if build_dir.exists():
            shutil.rmtree(docs_dir / 'build')
        os.system(f'(cd {docs_dir} && make html)')
    os.system(f'(cd {build_dir} && python -m http.server)')


@click.command()
@click.argument('group')
@click.option('--reset/--no-reset', default=False, help='Delete existing group keys in config')
def configure(group, reset):
    """Update configuration keys.

    Specify one of the following group of keys to update:
    \n\thf              : Add HuggingFace auth token
    \n\ttrain-args      : Set default arguments for training routines
    """

    config.update(group, reset)


@click.command()
@click.argument('data-path')
@click.argument('output-dir')
@click.argument('task')
@click.option('--model-name', default=config['MODEL_NAME'], help='Name of model to finetune')
@click.option('--loss', default=None, help='Name of loss to use for training')
@click.option('--metric', default=None, help='Name of metric to use for evaluation')
@click.option('--max-length', default=config['MAX_LENGTH'], help='Truncate inputs to max token sequence length')
@click.option('--eval-split', default=config['EVAL_SPLIT'], help='Proportion of data to use for evaluation')
@click.option('--epochs', default=config['EPOCHS'], help='Number of training epochs')
@click.option('--train-batch-size', default=config['TRAIN_BATCH_SIZE'], help='Train batch size')
@click.option('--eval-batch-size', default=config['EVAL_BATCH_SIZE'], help='Evaluation batch size')
@click.option('--gradient-accumulation-steps', default=config['GRADIENT_ACCUMULATION_STEPS'], help='Artificially increase the train batch size')
@click.option('--learning-rate', default=config['LEARNING_RATE'], help='Learning rate')
@click.option('--warmup-ratio', default=config['WARMUP_RATIO'], help='Learning rate warmup')
@click.option('--weight-decay', default=config['WEIGHT_DECAY'], help='Weight decay for AdamW')
@click.option('--save-steps', default=config['SAVE_STEPS'], help='Save model checkpoint every n steps')
@click.option('--push', default=None, help='model id to push to hub')
@click.option('--overwrite/--no-overwrite', default=False, help='Overwrite output dir if it exists')
@click.option('--multi-label-delimiter', default='|', help='Delimiter for splitting labels in multi-label-classification task')
@click.option('--text-col', default='text', help='The column with text')
@click.option('--label-col', default='label', help='The column with labels')
def train(
        data_path, output_dir, task, model_name, 
        loss, metric, max_length, eval_split, epochs, 
        train_batch_size, eval_batch_size, gradient_accumulation_steps,
        learning_rate, warmup_ratio, weight_decay,
        save_steps, push, overwrite, 
        multi_label_delimiter, text_col, label_col
    ):
    """
    Run a training routine from a csv file.

    Specify one of the following tasks:
    \n\tclassification
    \n\tmulti-label-classification
    \n\ttriplet-vector-encoding

    The format of the csv files should be as follows:

    ### classification

    """
    
    data = pd.read_csv(data_path)
    texts = list(data[text_col].values)
    labels = list(data[label_col].astype(str).values)
    if task == 'multi-label-classification':
        labels = [[x for x in labels.split(multi_label_delimiter) if x != ''] for labels in labels]
    
    routine = scales_nlp.training_routine(task,
        model_name=model_name, loss=loss, metric=metric, 
        max_length=max_length, eval_split=eval_split, epochs=epochs,
        train_batch_size=train_batch_size, eval_batch_size=eval_batch_size, 
        gradient_accumulation_steps=gradient_accumulation_steps, learning_rate=learning_rate,
        warmup_ratio=warmup_ratio, weight_decay=weight_decay,
        save_steps=save_steps,
    )

    routine.train(output_dir, texts, labels, push=push, overwrite=overwrite)


@click.group()
def main():
    """SCALES-NLP: An AI Toolkit for Legal Research"""
    pass


main.add_command(configure)
main.add_command(train)


if config['DEVELOPER_MODE']:
    main.add_command(build)
    main.add_command(docs)


if __name__=='__main__':
    main()