import requests 
import pandas as pd
import numpy as np
import torch
from toolz import partition_all, interleave
from sklearn.feature_extraction.text import TfidfVectorizer
import faiss
from tqdm import tqdm
from scales_nlp import config


class PairedSimilarExamplesDataset(torch.utils.data.IterableDataset):
    def __init__(self, tokenizer, search_index, texts, labels=None, shuffle=False, topk=100, max_length=512):
        self.tokenizer = tokenizer
        self.search_index = search_index
        self.texts = texts
        self.labels = labels
        self.shuffle = shuffle
        self.topk = topk
        self.max_length = max_length

    def __iter__(self):
        idxs = list(range(len(self.search_index)))
        if self.shuffle:
            np.random.shuffle(idxs)
            
        for batch in partition_all(1024, idxs):
            batch = list(batch)
            query = self.search_index[batch]
            results = self.search_index.search(query, k=self.topk + 1)
            results = [np.random.choice(x[1:]) for x in results]
            indices = batch + [x['index'] for x in results]
            texts = [self.texts[i] for i in indices]
            inputs = self.tokenizer(
                texts, 
                padding="max_length", 
                max_length=self.max_length, 
                truncation=True,
            )
            if self.labels is not None:
                inputs['label'] = [self.labels[i] for i in indices]
                if isinstance(inputs['label'][0], str):
                    inputs['label'] = [self.tokenizer(x, padding='max_length', truncation=True)['input_ids'] for x in inputs['label']]

            ordering = interleave([range(len(texts) // 2), range(len(texts) // 2, len(texts))])
            for i in ordering:
                yield {k: v[i] for k, v in inputs.items()}
    
    def __len__(self):
        return len(self.search_index) * 2


class PairedSimilarExamplesDataset2(torch.utils.data.IterableDataset):
    def __init__(self, texts, labels, tokenizer, max_length, shuffle=True):
        super(PairedSimilarExamplesDataset).__init__()
        self.embeddings = Embeddings(texts=texts, vectors=labels)
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.shuffle = shuffle

    def __len__(self):
        return len(self.embeddings.texts) * 2

    def __iter__(self):
        idxs = list(range(len(self.embeddings.texts)))
        if self.shuffle:
            np.random.shuffle(idxs)
            
        for idx in partition_all(128, idxs):
            idx = list(idx)
            query = self.embeddings.vectors[idx]
            results = self.embeddings.search(query, k=6)
            results = [np.random.choice(x[1:]) for x in results]
            texts = [self.embeddings.texts[i] for i in idx] + [x['text'] for x in results]
            idx = idx + [x['index'] for x in results]
            inputs = self.tokenizer(texts, 
                padding="max_length", 
                max_length=self.max_length, 
                truncation=True,
            )
            inputs['label'] = self.embeddings.vectors[idx]
            ordering = interleave([range(len(texts) // 2), range(len(texts) // 2, len(texts))])
            for i in ordering:
                yield {k: v[i] for k, v in inputs.items()}


class TokenClassificationDataset(torch.utils.data.Dataset):
    def __init__(self, tokenizer, texts, spans, label_names, max_length=512):
        self.tokenizer = tokenizer
        self.texts = texts
        self.spans = spans
        self.label_names = label_names
        self.label_map = {'O': 0}
        for label_name in sorted(label_names):
            self.label_map['B-' + label_name] = len(self.label_map)
            self.label_map['I-' + label_name] = len(self.label_map)
        self.max_length = max_length

    def __getitem__(self, i):
        text = self.texts[i]
        inputs = self.tokenizer(text, padding="max_length", max_length=self.max_length, truncation=True, return_tensors="pt", return_offsets_mapping=True)
        example = {k: v[0] for k, v in inputs.items()}
        example['labels'] = self.example_spans_to_labels(example, self.spans[i])
        del example['offset_mapping']
        return example

    def __len__(self):
        return len(self.texts)

    def example_spans_to_labels(self, example, spans):
        input_ids = example['input_ids']
        spans = sorted(spans, key=lambda x: x['start'])

        labels = []
        current_label = None
        for i in range(len(example['input_ids'])):
            offset = example['offset_mapping'][i]
            
            if len(spans) > 0 and offset[0] >= spans[0]['end']:
                spans.pop(0)
                current_label = None

            if offset[1] == 0:
                labels.append(-100)
                current_label = None
            elif len(spans) == 0 or offset[1] <= spans[0]['start']:
                labels.append(0)
                current_label = None
            else:
                if current_label is None:
                    current_label = spans[0]['label']
                    labels.append(self.label_map['B-' + current_label])
                else:
                    labels.append(self.label_map['I-' + current_label])

        return torch.tensor(labels)



