from collections.abc import Callable
from pathlib import Path
from typing import Union, Optional, List
import faiss
import numpy as np
import pandas as pd
from toolz import partition_all
from tqdm import tqdm


class SearchIndex():
    """
        A simple FAISS search index with stored texts

        :param texts: a list of strings
        :param encodings: a numpy array of encodings, should have shape (len(texts), encoding_dim)
        :param add_encodings: whether to add the encodings and text to the index - if False, encodings and text will only be used to initialize the index
    """
    def __init__(self, texts: List[str], encodings: np.ndarray, add_encodings: bool = True):
        self.texts = []
        self.index = faiss.IndexFlatL2(encodings.shape[-1])
        if add_encodings:
            self.texts = texts
            self.index.add(encodings.astype('float32'))
    
    def add(self, texts: List[str], encodings: np.ndarray):
        """
        Add new texts and encodings to the index

        :param texts: a list of strings
        :param encodings: a numpy array of encodings, should have shape (len(texts), encoding_dim)
        """
        self.texts += texts
        self.index.add(encodings.astype('float32'))

    def search(self, query: np.ndarray, k: int = 1) -> list[list[dict]]:
        """
        Search the index for the nearest neighbors of the query

        :param query: a numpy array of query encodings with shape (num_queries, encoding_dim)
        :param k: the number of results to return for each query row

        :return: a list of results, each result includes a list of k matches including the text, distance, and row index
        """
        index = self.index
        query = query.astype('float32')
        distances, indices = index.search(query, k=k)
        results: list[list[dict]] = []
        for i in range(len(query)):
            results.append([])
            for j in range(k):
                results[-1].append({
                    'text': self.texts[indices[i][j]],
                    'index': indices[i][j],
                    'distance': distances[i][j],
                })
        return results

    def save(self, output_dir: Union[str, Path]):
        """
        Save the index to disk

        :param output_dir: the directory to save the index to
        """
        index = self.index
        output_dir = Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        faiss.write_index(index, str(output_dir / 'index.faiss'))
        pd.DataFrame(self.texts, columns=['texts']).to_csv(output_dir / 'texts.csv', index=False)

    @staticmethod
    def load(input_dir: Union[str, Path]) -> 'SearchIndex':
        """
        Load a saved index from disk

        :param input_dir: the directory to load the index from
        :return: a search index
        """
        input_dir = Path(input_dir)
        index = faiss.read_index(str(input_dir / 'index.faiss'))
        search_index = SearchIndex([], encodings=np.zeros((0, index.d)), add_encodings=False)
        texts = pd.read_csv(input_dir / 'texts.csv')['texts'].tolist()
        search_index.texts = texts
        search_index.index = index
        return search_index

    def __getitem__(self, item: Union[int, slice, list[int]]) -> np.ndarray:
        """
        Get reconstructed encodings from the index
        :param item: An integer, slice, or list of integers
        :return: The text(s) at the specified index(es)
        """
        index = self.index
        if isinstance(item, int):
            item = int(item)
            if item < 0:
                item = len(self) + item
            return index.reconstruct(item).astype('float32')
        elif isinstance(item, slice):
            return np.vstack([self[i] for i in list(range(len(self))[item])]).astype('float32')
        elif isinstance(item, list):
            return np.vstack([self[i] for i in item]).astype('float32')
        else:
            return index.reconstruct(int(item)).astype('float32')

    def __len__(self):
        return len(self.texts)


class QuantizedSearchIndex(SearchIndex):
    pass
