import json
from pathlib import Path
from typing import Union, List, Tuple
import pandas as pd
import requests
import scales_nlp
from scales_nlp import config

PACKAGE_DIR = Path(__file__).parent
PACKAGE_DATA_DIR = PACKAGE_DIR / 'data'

LABEL_DATA_DIR = config['LABEL_DATA_DIR'] if config['LABEL_DATA_DIR'] is not None else config['PACER_DIR']

COURTS = pd.read_csv(PACKAGE_DATA_DIR / 'courts.csv')
STATES = COURTS['state'].dropna().unique()
DIVISIONS = COURTS['cardinal'].dropna().unique()


def courts() -> pd.DataFrame:
    return COURTS.copy()


def states() -> List[str]:
    return STATES


def divisions() -> List[str]:
    return DIVISIONS


def load_json(path: Union[str, Path]) -> dict:
    with open(str(path), 'r') as f:
        return json.loads(f.read())


def get_ucid_components(ucid: str) -> Tuple[str, str, str]:
    court = ucid.split(';;')[0]
    docket_number = ucid.split(';;')[1]
    office_number = docket_number.split(':')[0]
    year = docket_number.split(':')[1].split('-')[0]
    return court, docket_number, office_number, year


def case_path(ucid: str) -> Path:
    court, docket_number, _, year = get_ucid_components(ucid)
    filename = docket_number.replace(':', '-') + '.json'
    return config['PACER_DIR'] / court / 'json' / year / filename


def load_case(ucid: str) -> dict:
    path = case_path(ucid)
    return load_json(path)


def load_case_classifier_labels(ucid: str) -> List:
    court, docket_number, _, year = get_ucid_components(ucid)
    filename = docket_number.replace(':', '-') + '.json'
    path = LABEL_DATA_DIR / court / 'labels' / year / filename
    if path.exists():
        return load_json(path)
    else:
        print('labels not computed for {}'.format(ucid))
        return []


def load_court(court: str) -> dict:
    courts = COURTS[COURTS['abbreviation'] == court]
    if len(courts) == 0:
        raise ValueError(f'Court {court} not found')
    return courts.iloc[0].to_dict()


def convert_default_binary_outputs(
    predictions: List[Union[str, dict[str, float], bool]]
) -> List[Union[float, bool]]:
    if isinstance(predictions[0], str):
        converted_predictions = [prediction == 'LABEL_1' for prediction in predictions]
    elif isinstance(predictions[0], dict):
        converted_predictions = [prediction['LABEL_1'] for prediction in predictions]
    else:
        converted_predictions = predictions
    return converted_predictions


def get_label_studio_data(project_id:int) -> List[dict]:
    url = f"{scales_nlp.config['LABEL_STUDIO_URL']}/api/projects/{project_id}/export?exportType=JSON"
    headers = {'Authorization': f"Token {scales_nlp.config['LABEL_STUDIO_TOKEN']}"}

    r = requests.get(url, headers=headers)
    return r.json()


def get_label_studio_ner_data(project_id:int) -> Tuple[List[str], List[List[dict]]]:
    label_studio_data = get_label_studio_data(project_id)
    data = []
    for example in label_studio_data:
        spans = []
        for anno in example['annotations']:
            for result in anno['result']:
                value = result['value']
                spans.append({
                    'start': value['start'],
                    'end': value['end'],
                    'label': value['labels'][0],
                })

        data.append({'text': example['data']['text'], 'spans': spans})
    data = pd.DataFrame(data)
    return data['text'].tolist(), data['spans'].tolist()
